<?php

declare(strict_types=1);

namespace Drupal\flysystem\Entity;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Config\Config;
use Drupal\Core\Config\Entity\ConfigEntityBase;
use Drupal\Core\Config\Entity\ConfigEntityTypeInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\flysystem\FlysystemAdapterConfigInterface;
use Drupal\flysystem\FlysystemException;

/**
 * Defines the flysystem_adapter_config entity type.
 *
 * @ConfigEntityType(
 *   id = "flysystem_adapter_config",
 *   label = @Translation("Configured Flysystem Adapter"),
 *   label_collection = @Translation("Configured Flysystem Adapters"),
 *   label_singular = @Translation("configured Flysystem Adapter"),
 *   label_plural = @Translation("configured Flysystem Adapters"),
 *   label_count = @PluralTranslation(
 *     singular = "@count configured Flysystem Adapter",
 *     plural = "@count configured Flysystem Adapters",
 *   ),
 *   handlers = {
 *     "list_builder" = "Drupal\flysystem\FlysystemAdapterConfigListBuilder",
 *     "form" = {
 *       "add" = "Drupal\flysystem\Form\FlysystemAdapterConfigForm",
 *       "edit" = "Drupal\flysystem\Form\FlysystemAdapterConfigForm",
 *       "delete" = "Drupal\Core\Entity\EntityDeleteForm",
 *     },
 *   },
 *   config_prefix = "flysystem_adapter_config",
 *   admin_permission = "administer flysystem_adapter_config",
 *   links = {
 *     "collection" = "/admin/config/media/flysystem-adapter-config",
 *     "add-form" = "/admin/config/media/flysystem-adapter-config/add",
 *     "edit-form" = "/admin/config/media/flysystem-adapter-config/{flysystem_adapter_config}",
 *     "delete-form" = "/admin/config/media/flysystem-adapter-config/{flysystem_adapter_config}/delete",
 *   },
 *   entity_keys = {
 *     "id" = "id",
 *     "label" = "label",
 *     "uuid" = "uuid",
 *   },
 *   config_export = {
 *     "id",
 *     "label",
 *     "description",
 *     "sub_adapter",
 *     "sub_adapter_config",
 *   },
 * )
 */
final class FlysystemAdapterConfig extends ConfigEntityBase implements FlysystemAdapterConfigInterface {

  /**
   * The flysystem_adapter_config ID.
   */
  protected string $id;

  /**
   * The flysystem_adapter_config label.
   */
  protected string $label;

  /**
   * The flysystem_adapter_config description.
   */
  protected string $description;

  /**
   * The ID of the subAdapter plugin.
   *
   * @var string
   */
  protected $sub_adapter;

  /**
   * The adapter plugin configuration.
   *
   * @var array
   */
  protected $sub_adapter_config = [];

  /**
   * The adapter plugin instance.
   *
   * @var \Drupal\flysystem\FlysystemAdapterInterface|null
   */
  protected $subAdapterPlugin;

  /**
   * {@inheritdoc}
   */
  public function getDescription() {
    return !empty($this->description) ? $this->description : '';
  }

  /**
   * {@inheritdoc}
   */
  public function hasValidSubAdapter() {
    $subadapter_plugin_definition = \Drupal::service('plugin.manager.flysystem_adapter')->getDefinition($this->getSubAdapterId(), FALSE);
    return !empty($subadapter_plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public function getSubAdapterId() {
    return $this->sub_adapter;
  }

  /**
   * {@inheritdoc}
   */
  public function getSubAdapterConfigItem($key) {
    if (is_array($key)) {
      return NestedArray::getValue($this->sub_adapter_config, $key);
    }
    return $this->sub_adapter_config[$key] ?? '';
  }

  /**
   * {@inheritdoc}
   */
  public function getSubAdapter() {
    if (!$this->subAdapterPlugin) {
      $subadapter_plugin_manager = \Drupal::service('plugin.manager.flysystem_adapter');
      $config = $this->sub_adapter_config;
      $config['#flysystem_adapter_config'] = $this;
      if (!($this->subAdapterPlugin = $subadapter_plugin_manager->createInstance($this->getSubAdapterId(), $config))) {
        $sub_adapter_id = $this->getSubAdapterId();
        throw new FlysystemException("The adapter with ID '$sub_adapter_id' could not be retrieved.");
      }
    }
    return $this->subAdapterPlugin;
  }

  /**
   * {@inheritdoc}
   */
  public function getSubAdapterConfig() {
    return $this->sub_adapter_config;
  }

  /**
   * {@inheritdoc}
   */
  public function setSubAdapterConfig(array $sub_adapter_config) {
    $this->sub_adapter_config = $sub_adapter_config;
    // In case the subadapter plugin is already loaded, make sure the
    // configuration stays in sync.
    if ($this->subAdapterPlugin
        && $this->getSubAdapter()->getConfiguration() !== $sub_adapter_config) {
      $this->getSubAdapter()->setConfiguration($sub_adapter_config);
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function preSave(EntityStorageInterface $storage) {
    parent::preSave($storage);

    // The rest of the code only applies to updates.
    if (!isset($this->original)) {
      return;
    }
    // Retrieve active config overrides for this adapter config entity.
    $overrides = $this->getConfigOverrides($this);

    // If there are overrides for the sub_adapter or its configuration, attempt
    // to apply them for the preUpdate() call.
    if (isset($overrides['sub_adapter']) || isset($overrides['sub_adapter_config'])) {
      $sub_adapter_config = $this->getSubAdapterConfig();
      if (isset($overrides['sub_adapter_config'])) {
        $sub_adapter_config = $overrides['sub_adapter_config'];
      }
      $sub_adapter_id = $this->getSubAdapterId();
      if (isset($overrides['sub_adapter'])) {
        $sub_adapter_id = $overrides['sub_adapter'];
      }
      $subadapter_plugin_manager = \Drupal::service('plugin.manager.flysystem_adapter');
      $sub_adapter_config['#flysystem_adapter_config'] = $this;
      if (!($subadapter_plugin_manager->createInstance($sub_adapter_id, $sub_adapter_config))) {
        throw new FlysystemException("The adapter with ID '$sub_adapter_id' could not be retrieved.");
      }
    }
  }

  /**
   * Implements the magic __clone() method.
   *
   * Prevents the adapter plugin instance from being cloned.
   */
  public function __clone() {
    $this->subAdapterPlugin = NULL;
  }

  /**
   * Retrieves all overridden property values for the given config entity.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The config entity to check for overrides.
   *
   * @return array
   *   An associative array mapping property names to their overridden values.
   */
  private function getConfigOverrides(EntityInterface $entity) {
    $entity_type = $entity->getEntityType();
    if (!($entity_type instanceof ConfigEntityTypeInterface)) {
      return [];
    }

    $config_key = $entity_type->getConfigPrefix() . '.' . $entity->id();
    $config = \Drupal::config($config_key);
    if (!$config->hasOverrides()) {
      return [];
    }

    return $this->collectOverrides($config, $config->get());
  }

  /**
   * Collects overrides from a config object.
   *
   * @param \Drupal\Core\Config\Config $config
   *   The config object.
   * @param array $values
   *   The array of values for the given $prefix.
   * @param array $overrides
   *   (optional) The overrides collected so far. Internal use only.
   * @param string $prefix
   *   (optional) The config key prefix for the current call level.  Internal
   *   use only.
   *
   * @return array
   *   An associative array mapping property names to their overridden values.
   */
  private function collectOverrides(Config $config, array $values, array $overrides = [], $prefix = '') {
    foreach ($values as $key => $value) {
      $key = "$prefix$key";
      if (!$config->hasOverrides($key)) {
        continue;
      }
      if (is_array($value)) {
        NestedArray::setValue($overrides, explode('.', $key), []);
        $overrides = $this->collectOverrides($config, $value, $overrides, "$key.");
      }
      else {
        NestedArray::setValue($overrides, explode('.', $key), $value);
      }
    }

    return $overrides;
  }

}
