<?php

namespace Drupal\flysystem\FlyStream;

use Drupal\Core\Config\CachedStorage;
use Drupal\Core\StreamWrapper\PhpStreamWrapperInterface;
use Drupal\flysystem\FlyStream\Exception\CouldNotDeleteFileException;
use Drupal\flysystem\FlyStream\Exception\CouldNotRemoveDirectoryException;
use Drupal\flysystem\FlyStream\Exception\DirectoryExistsException;
use Drupal\flysystem\FlyStream\Exception\DirectoryNotEmptyException;
use Drupal\flysystem\FlyStream\Exception\DirectoryNotFoundException;
use Drupal\flysystem\FlyStream\Exception\FileNotFoundException;
use Drupal\flysystem\FlyStream\Exception\InvalidStreamModeException;
use Drupal\flysystem\FlyStream\Exception\IsDirectoryException;
use Drupal\flysystem\FlyStream\Exception\IsNotDirectoryException;
use Drupal\flysystem\FlyStream\Exception\RootDirectoryException;
use Drupal\flysystem\FlyStream\Exception\StatFailedException;
use Drupal\flysystem\FlyStream\Exception\UnableToChangePermissionsException;
use Drupal\flysystem\FlyStream\Exception\UnableToCreateDirectoryException;
use Drupal\flysystem\FlyStream\Exception\UnableToReadException;
use Drupal\flysystem\FlyStream\Exception\UnableToWriteException;
use League\Flysystem\Config;
use League\Flysystem\Filesystem;
use League\Flysystem\FilesystemException;
use League\Flysystem\StorageAttributes;
use League\Flysystem\UnableToRetrieveMetadata;
use League\Flysystem\UnixVisibility\PortableVisibilityConverter;
use League\Flysystem\Visibility;
use League\Flysystem\WhitespacePathNormalizer;
use Symfony\Component\Lock\Key;
use Symfony\Component\Lock\Lock;
use Symfony\Component\Lock\Store\StoreFactory;

/**
 * FlyStreamWrapper class, stream wrapper base for Flysystem Drupal module.
 */
class FlyStreamWrapperBase implements PhpStreamWrapperInterface, FlyStreamWrapperInterface {

  use ExceptionHandlerTrait;

  /**
   * Config Storage service.
   *
   * @var \Drupal\Core\Config\CachedStorage
   */
  protected $configStorage;
  /**
   * Configuration file name.
   *
   * @var string|null
   */
  protected $configuration;

  /**
   * File permission mode.
   *
   * @var string|null
   */
  protected $mode = NULL;

  /**
   * Directory listing.
   *
   * @var \Iterator|\IteratorAggregate|\Traversable|null
   */
  protected $dir = NULL;

  /**
   * Resource context.
   *
   * @var resource
   */
  public $context;

  /**
   * Directory Listing contents.
   *
   * @var \Iterator<mixed,\League\Flysystem\StorageAttributes>
   */
  public $dirListing;

  /**
   * File directory path.
   *
   * @var string
   */
  public $path;

  /**
   * Wrapper scheme.
   *
   * @var string
   */
  protected $scheme;

  /**
   * File path.
   *
   * @var string
   */
  public $file;

  /**
   * Flysystem Filesystem instance.
   *
   * @var \League\Flysystem\FilesystemOperator
   */
  public $filesystem;

  /**
   * Configuration data.
   *
   * @var array<string, int|string|bool|null>
   */
  public $config = [];

  /**
   * Work on Local Copy setting.
   *
   * @var bool
   */
  public $workOnLocalCopy = FALSE;

  /**
   * Bytes written to stream.
   *
   * @var int
   */
  public $bytesWritten = 0;

  /**
   * Lock Key setting.
   *
   * @var \Symfony\Component\Lock\Key
   */
  public $lockKey;

  /**
   * Write only setting.
   *
   * @var bool
   */
  public $writeOnly = FALSE;

  /**
   * Always Append setting.
   *
   * @var bool
   */
  public $alwaysAppend = FALSE;

  /**
   * Write Buffer Size.
   *
   * @var int
   */
  public $writeBufferSize = 0;

  /**
   * Resource handle.
   *
   * @var resource|false
   */
  public $handle = FALSE;

  /**
   * Adapter Factory options.
   *
   * @var array|null
   */
  public $options = NULL;

  /**
   * Class constructor.
   *
   * @param string $configuration
   *   Name of configuration file for adapter.
   * @param \Drupal\Core\Config\CachedStorage $config_storage
   *   Config Storage service.
   */
  public function __construct(string $configuration, CachedStorage $config_storage) {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    \Drupal::logger('FlyStreamWrapper')->notice('Configuration: ' . $configuration);
    \Drupal::logger('FlyStreamWrapper')->notice('config.storage ' . isset($config_storage));

    $this->configuration = $configuration;
    $this->configStorage = $config_storage;

    $config = $this->configStorage->read($configuration);

    $adapterClass = new \ReflectionClass($config['sub_adapter_config']['plugin_adapter_class']);

    $adapterFactory = $adapterClass->newInstance();

    $adapterFactoryInstance = $adapterFactory::create($config['sub_adapter_config']);

    $this->options = $adapterFactoryInstance['options'];
    $this->filesystem = new Filesystem($adapterFactoryInstance['adapter']);

  }

  // phpcs:disable Drupal.NamingConventions.ValidFunctionName.ScopeNotCamelCaps

  /**
   * {@inheritdoc}
   *
   * @todo Done
   */
  public function dir_closedir(): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    unset($this->dirListing);
    return TRUE;
  }

  /**
   * {@inheritdoc}
   *
   * @todo Done
   */
  public function dir_opendir($path, $options): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    $this->setPath($path);
    try {
      $listing = $this->filesystem->listContents($this->file)->getIterator();
      $this->dirListing = ($listing instanceof \Iterator) ? $listing : new \IteratorIterator($listing);
    }
    catch (FilesystemException $e) {
      return $this->triggerError(
        DirectoryNotFoundException::atLocation(__METHOD__, $path, $e)
      );
    }

    $valid = @is_dir($path);
    if (!$valid) {
      return self::triggerError(
        DirectoryNotFoundException::atLocation(__METHOD__, $path)
      );
    }

    return TRUE;

  }

  /**
   * {@inheritdoc}
   *
   * @todo Done
   */
  public function dir_readdir(): string|bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    if (!$this->dirListing->valid()) {
      return FALSE;
    }

    $item = $this->dirListing->current();

    $this->dirListing->next();

    return basename($item->path());

  }

  /**
   * {@inheritdoc}
   *
   * @todo Done
   */
  public function dir_rewinddir(): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    try {
      $listing = $this->filesystem->listContents($this->file)->getIterator();
      $this->dirListing = ($listing instanceof \Iterator) ? $listing : new \IteratorIterator($listing);
    }
    catch (FilesystemException $e) {
      return self::triggerError($e);
    }

    return TRUE;

  }

  /**
   * {@inheritdoc}
   *
   * @todo Done
   */
  public function mkdir($path, $mode, $options): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    if (file_exists($path)) {
      return self::triggerError(DirectoryExistsException::atLocation(__METHOD__, $path));
    }

    $this->setPath($path);

    try {
      $visibility = new PortableVisibilityConverter();
      $config = [
        Config::OPTION_VISIBILITY => $visibility->inverseForDirectory($mode),
      ];
      $this->filesystem->createDirectory($this->file, $config);

      return TRUE;
    }
    catch (FilesystemException $e) {
      return self::triggerError(
        UnableToCreateDirectoryException::atLocation(__METHOD__, $path, $e)
      );
    }
  }

  /**
   * {@inheritdoc}
   *
   * @todo Done
   */
  public function rename($path_from, $path_to): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    $this->setPath($path_from);
    $errorLocation = $path_from . ',' . $path_to;
    if (!file_exists($path_from)) {
      return self::triggerError(FileNotFoundException::atLocation(__METHOD__, $errorLocation));
    }

    if (file_exists($path_to)) {
      if (is_file($path_from) && is_dir($path_to)) {
        return self::triggerError(
          IsDirectoryException::atLocation(__METHOD__, $errorLocation)
        );
      }
      if (is_dir($path_from) && is_file($path_to)) {
        return self::triggerError(
          IsNotDirectoryException::atLocation(__METHOD__, $errorLocation)
        );
      }
    }

    try {
      $this->filesystem->move($this->file, $this->getFile($path_to));
      return TRUE;
    }
    catch (FilesystemException $e) {
      return self::triggerError(
        DirectoryNotEmptyException::atLocation(__METHOD__, $errorLocation, $e)
      );
    }
  }

  /**
   * {@inheritdoc}
   *
   * @todo Done
   */
  public function rmdir($path, $options): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    $this->setPath($path);

    $n = new WhitespacePathNormalizer();
    $n->normalizePath($this->file);
    if ('' === $n->normalizePath($this->file)) {
      return self::triggerError(
        RootDirectoryException::atLocation(__METHOD__, $this->path)
      );
    }

    if (($options & STREAM_MKDIR_RECURSIVE) !== 0) {
      try {
        $this->filesystem->deleteDirectory($this->file);
        return TRUE;
      }
      catch (FilesystemException $e) {
        return self::triggerError(
          CouldNotRemoveDirectoryException::atLocation(__METHOD__, $this->path, $e)
        );
      }
    }

    try {
      $listing = $this->filesystem->listContents($this->file);
    }
    catch (FilesystemException $e) {
      return self::triggerError(
        DirectoryNotEmptyException::atLocation(__METHOD__, $this->path)
      );
    }

    foreach ($listing as $ignored) {
      if (!$ignored instanceof StorageAttributes) {
        continue;
      }

      return self::triggerError(
        DirectoryNotEmptyException::atLocation(__METHOD__, $this->path)
      );
    }

    try {
      $this->filesystem->deleteDirectory($this->file);
      return TRUE;
    }
    catch (FilesystemException $e) {
      return self::triggerError(
        CouldNotRemoveDirectoryException::atLocation(__METHOD__, $this->path, $e)
      );
    }
  }

  /**
   * {@inheritdoc}
   *
   * @todo Done
   */
  public function stream_cast($cast_as) {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    return $this->handle;
  }

  /**
   * {@inheritdoc}
   *
   * @todo Done
   */
  public function stream_close(): void {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    if (!is_resource($this->handle)) {
      return;
    }

    if ($this->workOnLocalCopy) {
      fflush($this->handle);
      rewind($this->handle);

      try {
        $this->filesystem->writeStream($this->file, $this->handle);
      }
      catch (FilesystemException $e) {
        trigger_error(
          'stream_close(' . $this->path . ') Unable to sync file : ' . $e->getMessage(),
          E_USER_WARNING
        );
      }
    }
    fclose($this->handle);
  }

  /**
   * {@inheritdoc}
   *
   * @todo Done
   */
  public function stream_eof(): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    if (!is_resource($this->handle)) {
      return FALSE;
    }

    return feof($this->handle);
  }

  /**
   * {@inheritdoc}
   *
   * @todo Done
   */
  public function stream_flush(): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    if (!is_resource($this->handle)) {
      trigger_error(
        'stream_flush(): Supplied resource is not a valid stream resource',
        E_USER_WARNING
      );

      return FALSE;
    }

    $success = fflush($this->handle);

    if ($this->workOnLocalCopy) {
      fflush($this->handle);
      $currentPosition = ftell($this->handle);
      rewind($this->handle);

      try {
        $this->filesystem->writeStream($this->file, $this->handle);
      }
      catch (FilesystemException $e) {
        trigger_error(
          'stream_flush(' . $this->path . ') Unable to sync file : ' . $e->getMessage(),
          E_USER_WARNING
        );
        $success = FALSE;
      }

      if (FALSE !== $currentPosition) {
        if (is_resource($this->handle)) {
          fseek($this->handle, $currentPosition);
        }
      }
    }

    $this->bytesWritten = 0;

    return $success;
  }

  /**
   * {@inheritdoc}
   *
   * @todo In progress
   */
  public function stream_lock($operation): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    if (NULL === $this->lockKey) {
      $this->lockKey = new Key($this->path);
    }

    $store = StoreFactory::createStore((string) $this->config[FlyStreamWrapperInterface::LOCK_STORE]);
    $lock = new Lock(
      $this->lockKey,
      $store,
      (float) $this->config[FlyStreamWrapperInterface::LOCK_TTL],
      FALSE
    );

    switch ($operation) {
      case LOCK_SH:
        return $lock->acquireRead(TRUE);

      case LOCK_EX:
        return $lock->acquire(TRUE);

      case LOCK_UN:
        $lock->release();
        return TRUE;

      case LOCK_SH | LOCK_NB:
        return $lock->acquireRead();

      case LOCK_EX | LOCK_NB:
        return $lock->acquire();
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   *
   * @todo Done
   */
  public function stream_metadata($path, $option, $value): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    $this->setPath($path);
    $filesystem = $this->filesystem;
    $file = $this->file;
    $converter = $this->options['visibility'] ?? new PortableVisibilityConverter();
    switch ($option) {
      case STREAM_META_ACCESS:
        if (!is_int($value)) {
          $value = (int) $value;
        }
        $visibility = is_dir($path) ? $converter->inverseForDirectory($value) : $converter->inverseForFile($value);

        try {
          $filesystem->setVisibility($file, $visibility);
        }
        catch (FilesystemException $e) {
          if (!$this->ignoreVisibilityErrors()) {
            return self::triggerError(UnableToChangePermissionsException::atLocation(
              __METHOD__,
              $this->path,
              decoct($value),
              $e
            ));
          }
        }
        return TRUE;

      case STREAM_META_TOUCH:
        try {
          if (!$filesystem->fileExists($file)) {
            $filesystem->write($file, '');
          }
        }
        catch (FilesystemException $e) {
          return self::triggerError(UnableToWriteException::atLocation(
            __METHOD__,
            $this->path,
            $e
          ));
        }

        return TRUE;

      default:
        return FALSE;
    }

  }

  /**
   * {@inheritdoc}
   *
   * @todo Done
   */
  public function stream_open($path, $mode, $options, &$openedPath = NULL): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    $this->setPath($path);
    $filesystem = $this->filesystem;
    $file = $this->file;

    if (!preg_match('/^[rwacx](\+b?|b\+?)?$/', $mode)) {
      return self::triggerError(InvalidStreamModeException::atLocation(
        __METHOD__,
        $this->path,
        $mode
      ));
    }

    $this->writeOnly = !strpos($mode, '+');
    try {
      if ('r' === $mode[0] && $this->writeOnly) {
        $this->handle = $filesystem->readStream($file);
        $this->workOnLocalCopy = FALSE;
        $this->writeOnly = FALSE;
      }
      else {
        $this->handle = fopen('php://temp', 'w+b');
        $this->workOnLocalCopy = TRUE;

        if ('w' !== $mode[0] && $filesystem->fileExists($file)) {
          if ('x' === $mode[0]) {
            throw UnableToWriteException::atLocation(__METHOD__, $this->path);
          }

          $result = FALSE;
          if (is_resource($this->handle)) {
            $result = stream_copy_to_stream($filesystem->readStream($file), $this->handle);
          }
          if (!$result) {
            throw UnableToWriteException::atLocation(__METHOD__, $this->path);
          }
        }
      }

      $this->alwaysAppend = 'a' === $mode[0];
      if (is_resource($this->handle) && !$this->alwaysAppend) {
        @rewind($this->handle);
      }
    }
    catch (FilesystemException $e) {
      if (($options & STREAM_REPORT_ERRORS) !== 0) {
        return self::triggerError(UnableToReadException::atLocation(
          __METHOD__,
          $this->path,
          $e
        ));
      }
      return FALSE;
    }

    if ($this->handle && $options & STREAM_USE_PATH) {
      $openedPath = $path;
    }

    if (is_resource($this->handle)) {
      return TRUE;
    }

    if (($options & STREAM_REPORT_ERRORS) !== 0) {
      return self::triggerError(FileNotFoundException::atLocation(__METHOD__, $this->path));
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   *
   * @todo DONE
   */
  public function stream_read($count): string {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    if ($this->writeOnly || !is_resource($this->handle) || $count < 0) {
      return '';
    }

    return (string) fread($this->handle, $count);
  }

  /**
   * {@inheritdoc}
   */
  public function stream_seek($offset, $whence = SEEK_SET): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    if (!is_resource($this->handle)) {
      return FALSE;
    }

    return 0 === fseek($this->handle, $offset, $whence);
  }

  /**
   * {@inheritdoc}
   */
  public function stream_set_option($option, $arg1, $arg2 = NULL): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    if (!is_resource($this->handle)) {
      return FALSE;
    }

    switch ($option) {
      case STREAM_OPTION_BLOCKING:
        return stream_set_blocking($this->handle, 1 === $arg1);

      case STREAM_OPTION_READ_BUFFER:
        return 0 === stream_set_read_buffer(
          $this->handle,
            STREAM_BUFFER_NONE === $arg1 ? 0 : (int) $arg2
          );

      case STREAM_OPTION_WRITE_BUFFER:
        $this->writeBufferSize = STREAM_BUFFER_NONE === $arg1 ? 0 : (int) $arg2;

        return TRUE;

      case STREAM_OPTION_READ_TIMEOUT:
        return stream_set_timeout($this->handle, $arg1, (int) $arg2);
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function stream_stat(): array|false {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    try {
      return $this->getStat();
    }
    catch (FilesystemException $e) {
      self::triggerError(StatFailedException::atLocation(__METHOD__, $this->path, $e));

      return FALSE;
    }

  }

  /**
   * {@inheritdoc}
   */
  public function stream_tell(): int {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    if (!is_resource($this->handle)) {
      return 0;
    }

    if ($this->alwaysAppend && $this->writeOnly) {
      return 0;
    }

    return (int) ftell($this->handle);
  }

  /**
   * {@inheritdoc}
   */
  public function stream_truncate($new_size): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    if (!is_resource($this->handle) || $new_size < 0) {
      return FALSE;
    }

    return ftruncate($this->handle, $new_size);
  }

  /**
   * {@inheritdoc}
   */
  public function stream_write($data): int|false {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    if (!is_resource($this->handle)) {
      return 0;
    }

    if ($this->alwaysAppend) {
      fseek($this->handle, 0, SEEK_END);
    }

    $size = (int) fwrite($this->handle, $data);
    $this->bytesWritten += $size;

    if ($this->alwaysAppend) {
      fseek($this->handle, 0, SEEK_SET);
    }

    return $size;
  }

  /**
   * {@inheritdoc}
   */
  public function unlink($path): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    $this->setPath($path);

    if (!file_exists($this->path)) {
      return self::triggerError(FileNotFoundException::atLocation(__METHOD__, $this->path));
    }

    try {
      $this->filesystem->delete($this->file);
      return TRUE;
    }
    catch (FilesystemException $e) {
      return self::triggerError(
        CouldNotDeleteFileException::atLocation(__METHOD__, $this->path, $e)
          );
    }
  }

  /**
   * {@inheritdoc}
   */
  public function url_stat($path, $flags): array|false {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    $this->setPath($path);

    try {
      return $this->getStat();
    }
    catch (FilesystemException $e) {
      if (($flags & STREAM_URL_STAT_QUIET) !== 0) {
        return FALSE;
      }

      self::triggerError(StatFailedException::atLocation(__METHOD__, $path, $e));

      return FALSE;
    }
  }

  // phpcs:enable

  /**
   * Get file, directory, or resource stat.
   *
   * @return array<int|string,int|string>|false
   *   Requested stat, or FALSE if it doesn't exist.
   *
   * @throws \League\Flysystem\FilesystemException
   */
  protected function getStat() {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    $stats = [];

    if ($this->workOnLocalCopy && is_resource($this->handle)) {
      $stats = fstat($this->handle);
      if (!$stats) {
        return FALSE;
      }
      if ($this->filesystem->fileExists($this->file)) {
        [$mode, $size, $time] = $this->getRemoteStats();

        unset($size);
      }
    }
    else {
      [$mode, $size, $time] = $this->getRemoteStats();
    }

    foreach (self::STATS_ZERO as $key) {
      $stats[$key] = 0;
    }

    foreach (self::STATS_MINUS_ONE as $key) {
      $stats[$key] = -1;
    }

    if (isset($mode)) {
      foreach (self::STATS_MODE as $key) {
        $stats[$key] = $mode;
      }
    }

    if (isset($size)) {
      foreach (self::STATS_SIZE as $key) {
        $stats[$key] = $size;
      }
    }

    if (isset($time)) {
      foreach (self::STATS_TIME as $key) {
        $stats[$key] = $time;
      }
    }

    $stats['uid'] = $stats[4] = (int) $this->config[FlyStreamWrapperInterface::UID];
    $stats['gid'] = $stats[5] = (int) $this->config[FlyStreamWrapperInterface::GID];

    return $stats;
  }

  /**
   * Gets stats from remote resource.
   *
   * @throws \League\Flysystem\FilesystemException
   *
   * @return array<int,int>
   *   Requested stats.
   */
  protected function getRemoteStats(): array {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    $newConverter = new PortableVisibilityConverter(
      (int) $this->config[FlyStreamWrapperInterface::VISIBILITY_FILE_PUBLIC],
      (int) $this->config[FlyStreamWrapperInterface::VISIBILITY_FILE_PRIVATE],
      (int) $this->config[FlyStreamWrapperInterface::VISIBILITY_DIRECTORY_PUBLIC],
      (int) $this->config[FlyStreamWrapperInterface::VISIBILITY_DIRECTORY_PRIVATE],
      (string) $this->config[FlyStreamWrapperInterface::VISIBILITY_DEFAULT_FOR_DIRECTORIES]
    );

    $converter = $this->options['visibility'] ?? $newConverter;
    try {
      $visibility = $this->filesystem->visibility($this->file);
    }
    catch (UnableToRetrieveMetadata | \TypeError $e) {
      if (!$this->ignoreVisibilityErrors()) {
        throw $e;
      }

      $visibility = Visibility::PUBLIC;
    }

    $mode = 0;
    $size = 0;
    $lastModified = 0;

    try {
      if ('directory' === $this->filesystem->mimeType($this->file)) {
        [$mode, $size, $lastModified] = $this->getRemoteDirectoryStats($converter, $visibility);
      }
      else {
        [$mode, $size, $lastModified] = $this->getRemoteFileStats($converter, $visibility);
      }
    }
    catch (UnableToRetrieveMetadata $e) {
      if (method_exists($this->filesystem, 'directoryExists')) {
        if ($this->filesystem->directoryExists($this->file)) {
          [$mode, $size, $lastModified] = $this->getRemoteDirectoryStats($converter, $visibility);
        }
        elseif ($this->filesystem->fileExists($this->file)) {
          [$mode, $size, $lastModified] = $this->getRemoteFileStats($converter, $visibility);
        }
      }
      else {
        throw $e;
      }
    }

    return [$mode, $size, $lastModified];
  }

  /**
   * Get stats for Remote Directory.
   *
   * @return array<int, int>
   *   Requested stats.
   *
   * @throws \League\Flysystem\FilesystemException
   */
  private function getRemoteDirectoryStats(
    PortableVisibilityConverter $converter,
    string $visibility,
  ): array {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    $mode = 040000 + $converter->forDirectory($visibility);
    $size = 0;
    $lastModified = $this->getRemoteDirectoryLastModified();
    return [$mode, $size, $lastModified];
  }

  /**
   * Get stats for remote file.
   *
   * @return array<int, int>
   *   Requested stats.
   *
   * @throws \League\Flysystem\FilesystemException
   */
  private function getRemoteFileStats(
    PortableVisibilityConverter $converter,
    string $visibility,
  ): array {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    $mode = 0100000 + $converter->forFile($visibility);
    $size = $this->filesystem->fileSize($this->file);
    $lastModified = $this->filesystem->lastModified($this->file);
    return [$mode, $size, $lastModified];
  }

  /**
   * Gets Last Modified stat for remote directory.
   *
   * @return int
   *   Last modified stat.
   *
   * @throws \League\Flysystem\FilesystemException
   */
  private function getRemoteDirectoryLastModified(): int {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    if (!$this->emulateDirectoryLastModified()) {
      return $this->filesystem->lastModified($this->file);
    }

    $lastModified = 0;
    $listing = $this->filesystem->listContents($this->file)->getIterator();
    $dirListing = $listing instanceof \Iterator ? $listing : new \IteratorIterator($listing);

    /** @var \League\Flysystem\FileAttributes $item */
    foreach ($dirListing as $item) {
      $lastModified = max($lastModified, $item->lastModified());
    }
    return $lastModified;
  }

  /**
   * Get the file path.
   *
   * @param string $path
   *   File path.
   *
   * @return string
   *   File path converted.
   */
  private function getFile(string $path): string {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    return (string) substr($path, strpos($path, '://') + 3);
  }

  /**
   * Configures settings based on path.
   *
   * @param string $path
   *   File path.
   */
  private function setPath(string $path): void {
    $this->path = $path;
    $this->scheme = substr($path, 0, (int) strpos($path, '://'));
    $this->file = $this->getFile($path);
  }

  /**
   * Gets Emulate Directory Last Modified settings.
   *
   * @return bool
   *   Boolean setting.
   */
  private function emulateDirectoryLastModified(): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    // @refactor??
    return (bool) $this->config[FlyStreamWrapper::EMULATE_DIRECTORY_LAST_MODIFIED];
  }

  /**
   * Gets Ignore Visibility Errors settings.
   *
   * @return bool
   *   Boolean setting.
   */
  private function ignoreVisibilityErrors(): bool {
    \Drupal::logger('FlyStreamWrapper')->notice('Calling ' . __METHOD__);
    return (bool) $this->config[FlyStreamWrapper::IGNORE_VISIBILITY_ERRORS];
  }

}
